import os
import sys

import xappt_qt
import xappt_qt.browser
import xappt_qt.launcher
import xappt_qt.builder


def main() -> int:
    argc = len(sys.argv)
    if argc == 1:
        return xappt_qt.browser.entry_point()
    else:
        if sys.argv[1] == "builder" and argc > 2:
            return xappt_qt.builder.main(sys.argv[2:])
        return xappt_qt.launcher.entry_point()


if __name__ == '__main__':
    if getattr(xappt_qt, "__compiled__", None) is not None:
        xappt_qt.executable = os.path.abspath(sys.argv[0])
    elif getattr(sys, "frozen", False) and hasattr(sys, "_MEIPASS"):
        xappt_qt.executable = os.path.abspath(sys.executable)
    sys.exit(main())



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'runner.ui'
#
# Created by: PyQt5 UI code generator 5.15.1
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_RunDialog(object):
    def setupUi(self, RunDialog):
        RunDialog.setObjectName("RunDialog")
        RunDialog.resize(753, 645)
        self.gridLayout_2 = QtWidgets.QGridLayout(RunDialog)
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.splitter = QtWidgets.QSplitter(RunDialog)
        self.splitter.setOrientation(QtCore.Qt.Vertical)
        self.splitter.setHandleWidth(12)
        self.splitter.setObjectName("splitter")
        self.widget = QtWidgets.QWidget(self.splitter)
        self.widget.setObjectName("widget")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.widget)
        self.verticalLayout_2.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.scrollArea = QtWidgets.QScrollArea(self.widget)
        self.scrollArea.setWidgetResizable(True)
        self.scrollArea.setObjectName("scrollArea")
        self.scrollAreaWidgetContents = QtWidgets.QWidget()
        self.scrollAreaWidgetContents.setGeometry(QtCore.QRect(0, 0, 731, 186))
        self.scrollAreaWidgetContents.setObjectName("scrollAreaWidgetContents")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.scrollAreaWidgetContents)
        self.verticalLayout.setObjectName("verticalLayout")
        self.gridLayout = QtWidgets.QGridLayout()
        self.gridLayout.setObjectName("gridLayout")
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.gridLayout.addItem(spacerItem, 1, 0, 1, 1)
        self.placeholder = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.placeholder.sizePolicy().hasHeightForWidth())
        self.placeholder.setSizePolicy(sizePolicy)
        self.placeholder.setAlignment(QtCore.Qt.AlignCenter)
        self.placeholder.setObjectName("placeholder")
        self.gridLayout.addWidget(self.placeholder, 0, 0, 1, 1)
        self.verticalLayout.addLayout(self.gridLayout)
        self.scrollArea.setWidget(self.scrollAreaWidgetContents)
        self.verticalLayout_2.addWidget(self.scrollArea)
        self.progressBar = QtWidgets.QProgressBar(self.widget)
        self.progressBar.setProperty("value", 0)
        self.progressBar.setTextVisible(True)
        self.progressBar.setFormat("")
        self.progressBar.setObjectName("progressBar")
        self.verticalLayout_2.addWidget(self.progressBar)
        self.txtOutput = QtWidgets.QTextEdit(self.splitter)
        self.txtOutput.setTabChangesFocus(True)
        self.txtOutput.setUndoRedoEnabled(False)
        self.txtOutput.setLineWrapMode(QtWidgets.QTextEdit.NoWrap)
        self.txtOutput.setReadOnly(True)
        self.txtOutput.setTabStopWidth(40)
        self.txtOutput.setObjectName("txtOutput")
        self.gridLayout_2.addWidget(self.splitter, 0, 0, 1, 1)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem1)
        self.btnOk = QtWidgets.QPushButton(RunDialog)
        self.btnOk.setObjectName("btnOk")
        self.horizontalLayout.addWidget(self.btnOk)
        self.btnClose = QtWidgets.QPushButton(RunDialog)
        self.btnClose.setObjectName("btnClose")
        self.horizontalLayout.addWidget(self.btnClose)
        spacerItem2 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem2)
        self.gridLayout_2.addLayout(self.horizontalLayout, 1, 0, 1, 1)

        self.retranslateUi(RunDialog)
        QtCore.QMetaObject.connectSlotsByName(RunDialog)

    def retranslateUi(self, RunDialog):
        _translate = QtCore.QCoreApplication.translate
        RunDialog.setWindowTitle(_translate("RunDialog", "Dialog"))
        self.placeholder.setText(_translate("RunDialog", "placeholder"))
        self.txtOutput.setHtml(_translate("RunDialog", "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0//EN\" \"http://www.w3.org/TR/REC-html40/strict.dtd\">\n"
"<html><head><meta name=\"qrichtext\" content=\"1\" /><style type=\"text/css\">\n"
"p, li { white-space: pre-wrap; }\n"
"</style></head><body style=\" font-family:\'Ubuntu\'; font-size:11pt; font-weight:400; font-style:normal;\" bgcolor=\"#000000\">\n"
"<p style=\"-qt-paragraph-type:empty; margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><br /></p></body></html>"))
        self.btnOk.setText(_translate("RunDialog", "Run"))
        self.btnClose.setText(_translate("RunDialog", "Close"))



import os
import platform
import sys


from PyQt5 import QtWidgets, QtGui, QtCore

import xappt

import xappt_qt.config
from xappt_qt.gui.ui.browser import Ui_Browser
from xappt_qt.gui.utilities.dark_palette import apply_palette
from xappt_qt.gui.utilities.tray_icon import TrayIcon
from xappt_qt.constants import *
from xappt_qt.gui.tab_pages import ToolsTabPage, OptionsTabPage, AboutTabPage

# noinspection PyUnresolvedReferences
from xappt_qt.gui.resources import icons

DISABLE_TRAY_ICON = platform.system() == "Darwin"


class XapptBrowser(xappt.ConfigMixin, QtWidgets.QMainWindow, Ui_Browser):
    def __init__(self):
        super().__init__()

        self.setupUi(self)
        self.setWindowIcon(QtGui.QIcon(":appicon"))
        self.setWindowTitle(APP_TITLE)

        self.tray_icon = TrayIcon(self, QtGui.QIcon(":appicon"))

        self.tools = ToolsTabPage(on_info=self.tray_icon.info, on_warn=self.tray_icon.warn,
                                  on_error=self.tray_icon.critical)
        self.options = OptionsTabPage()
        self.about = AboutTabPage()

        self.tabWidget.addTab(self.tools, self.tools.windowTitle())
        self.tabWidget.addTab(self.options, self.options.windowTitle())
        self.tabWidget.addTab(self.about, self.about.windowTitle())
        self.tabWidget.setCurrentIndex(0)

        self.config_path = APP_CONFIG_PATH.joinpath("browser.cfg")
        self.init_config()
        self.load_config()

        self.init_tray_icon()

    def init_config(self):
        self.add_config_item('launch-new-process',
                             saver=lambda: xappt_qt.config.launch_new_process,
                             loader=lambda x: setattr(xappt_qt.config, "launch_new_process", x),
                             default=True)
        self.add_config_item('minimize-to-tray',
                             saver=lambda: xappt_qt.config.minimize_to_tray,
                             loader=lambda x: setattr(xappt_qt.config, "minimize_to_tray", x),
                             default=True)
        self.add_config_item('window-size',
                             saver=lambda: (self.width(), self.height()),
                             loader=lambda x: self.setGeometry(0, 0, *x),
                             default=(350, 600))
        self.add_config_item('window-position',
                             saver=lambda: (self.geometry().x(), self.geometry().y()),
                             loader=lambda x: self.set_window_position(*x),
                             default=(-1, -1))
        if DISABLE_TRAY_ICON:
            self.options.chkMinimizeToTray.setChecked(False)
            self.options.chkMinimizeToTray.setEnabled(False)
            xappt_qt.config.minimize_to_tray = False

    def init_tray_icon(self):
        if DISABLE_TRAY_ICON:
            return

        for category, plugin_list in self.tools.loaded_plugins.items():
            for plugin in plugin_list:
                self.tray_icon.add_menu_item(plugin.name(), on_activate=self.run_tool, data=plugin, group=category)

        self.tray_icon.add_menu_item(None)
        self.tray_icon.add_menu_item("Show", on_activate=self.show, is_visible=self.isHidden)
        self.tray_icon.add_menu_item("Hide", on_activate=self.hide, is_visible=self.isVisible)
        self.tray_icon.add_menu_item(None)
        self.tray_icon.add_menu_item("Quit", on_activate=self.on_quit)
        self.tray_icon.on_trigger = self.on_activate
        self.tray_icon.show()

    def set_window_position(self, x: int, y: int):
        if x < 0 or y < 0:
            app = QtWidgets.QApplication.instance()
            cursor_pos = QtGui.QCursor.pos()
            screen = app.screenAt(cursor_pos)

            screen_rect = screen.availableGeometry()
            window_rect = QtCore.QRect(QtCore.QPoint(0, 0), self.frameSize().boundedTo(screen_rect.size()))
            self.resize(window_rect.size())
            self.move(screen_rect.center() - window_rect.center())
        else:
            self.move(x, y)

    def changeEvent(self, event: QtCore.QEvent):
        if not DISABLE_TRAY_ICON:
            if event.type() == QtCore.QEvent.WindowStateChange:
                if self.windowState() == QtCore.Qt.WindowMinimized:
                    if xappt_qt.config.minimize_to_tray and self.tray_icon.tray_available:
                        self.hide()
        super().changeEvent(event)

    def closeEvent(self, event: QtGui.QCloseEvent):
        if xappt_qt.config.minimize_to_tray and self.tray_icon.tray_available:
            event.ignore()
            self.hide()
        else:
            self.on_quit()

    def on_activate(self):
        self.show()
        self.raise_()
        self.activateWindow()

    def on_quit(self):
        self.tray_icon.destroy()
        self.save_config()
        QtWidgets.QApplication.instance().quit()

    def run_tool(self, **kwargs):
        plugin = kwargs.get('data')
        if plugin is None:
            return
        assert issubclass(plugin, xappt.BaseTool)
        self.tools.launch_tool(plugin)


def main(args) -> int:
    app = QtWidgets.QApplication.instance() or QtWidgets.QApplication(args)
    apply_palette(app)

    browser = XapptBrowser()
    browser.show()

    app.setProperty(APP_PROPERTY_RUNNING, True)
    return app.exec_()


def entry_point() -> int:
    os.environ[xappt.INTERFACE_ENV] = APP_INTERFACE_NAME
    return main(sys.argv)


if __name__ == '__main__':
    sys.exit(entry_point())



import argparse
import sys

from PyQt5 import QtWidgets
import xappt

from xappt_qt.gui.utilities.dark_palette import apply_palette
from xappt_qt.constants import *


def main(argv) -> int:
    parser = argparse.ArgumentParser()
    parser.add_argument('toolname', help='Specify the name of the tool to load')

    options = parser.parse_args(args=argv)

    xappt.discover_plugins()

    tool_class = xappt.get_tool_plugin(options.toolname)
    if tool_class is None:
        raise SystemExit(f"Tool {options.toolname} not found.")

    app = QtWidgets.QApplication.instance() or QtWidgets.QApplication(argv)
    apply_palette(app)

    app.setProperty(APP_PROPERTY_RUNNING, True)

    interface = xappt.get_interface()
    tool_instance = tool_class(interface=interface)
    interface.invoke(tool_instance)

    return app.exec_()


def entry_point() -> int:
    return main(sys.argv[1:])


if __name__ == '__main__':
    sys.exit(entry_point())



import platform
import subprocess
import sys

from PyQt5 import QtWidgets, QtGui, QtCore

from collections import defaultdict
from typing import DefaultDict, List, Tuple, Type

import xappt

import xappt_qt
import xappt_qt.config
from xappt_qt.constants import APP_TITLE
from xappt_qt.gui.ui.browser_tab_tools import Ui_tabTools
from xappt_qt.gui.delegates import SimpleItemDelegate
from xappt_qt.gui.tab_pages.base import BaseTabPage


class ToolsTabPage(BaseTabPage, Ui_tabTools):
    ROLE_TOOL_CLASS = QtCore.Qt.UserRole + 1
    ROLE_ITEM_TYPE = QtCore.Qt.UserRole + 2

    ITEM_TYPE_COLLECTION = 0
    ITEM_TYPE_TOOL = 1

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.setupUi(self)
        self.treeTools.setItemDelegate(SimpleItemDelegate())
        self.loaded_plugins: DefaultDict[str, List[Type[xappt.BaseTool]]] = defaultdict(list)
        self.populate_plugins()
        self.connect_signals()

    def populate_plugins(self):
        self.treeTools.clear()
        plugin_list: DefaultDict[str, List[Type[xappt.BaseTool]]] = defaultdict(list)

        for _, plugin_class in xappt.plugin_manager.registered_tools():
            collection = plugin_class.collection()
            plugin_list[collection].append(plugin_class)

        for collection in sorted(plugin_list.keys(), key=lambda x: x.lower()):
            collection_item = self._create_collection_item(collection)
            self.treeTools.insertTopLevelItem(self.treeTools.topLevelItemCount(), collection_item)
            for plugin in sorted(plugin_list[collection], key=lambda x: x.name().lower()):
                tool_item = self._create_tool_item(plugin)
                collection_item.addChild(tool_item)
                self.loaded_plugins[collection].append(plugin)
            collection_item.setExpanded(True)

    def connect_signals(self):
        self.treeTools.itemActivated.connect(self.item_activated)
        self.treeTools.itemSelectionChanged.connect(self.selection_changed)

        self.txtSearch.textChanged.connect(self.on_filter_tools)
        # noinspection PyAttributeOutsideInit
        self.__txtSearch_keyPressEvent_orig = self.txtSearch.keyPressEvent
        self.txtSearch.keyPressEvent = self._filter_key_press

    def _create_collection_item(self, collection_name: str) -> QtWidgets.QTreeWidgetItem:
        item = QtWidgets.QTreeWidgetItem()
        item.setText(0, collection_name)
        item.setData(0, self.ROLE_TOOL_CLASS, None)
        item.setData(0, self.ROLE_ITEM_TYPE, self.ITEM_TYPE_COLLECTION)
        return item

    def _create_tool_item(self, tool_class: Type[xappt.BaseTool]) -> QtWidgets.QTreeWidgetItem:
        item = QtWidgets.QTreeWidgetItem()
        item.setText(0, tool_class.name())
        item.setToolTip(0, tool_class.help())
        item.setData(0, self.ROLE_TOOL_CLASS, tool_class)
        item.setData(0, self.ROLE_ITEM_TYPE, self.ITEM_TYPE_TOOL)
        return item

    def item_activated(self, item: QtWidgets.QTreeWidgetItem, column: int):
        item_type = item.data(column, self.ROLE_ITEM_TYPE)
        if item_type != self.ITEM_TYPE_TOOL:
            return
        tool_class: Type[xappt.BaseTool] = item.data(column, self.ROLE_TOOL_CLASS)
        self.launch_tool(tool_class)

    @staticmethod
    def launch_command(tool_name: str) -> Tuple:
        if xappt_qt.executable is not None:
            return xappt_qt.executable, tool_name
        return sys.executable, "-m", "xappt_qt.launcher", tool_name

    def launch_tool(self, tool_class: Type[xappt.BaseTool]):
        if xappt_qt.config.launch_new_process:
            self.launch_tool_new_process(tool_class)
            return
        interface = xappt.get_interface()
        tool_instance = tool_class(interface=interface)
        interface.invoke(tool_instance)

    def launch_tool_new_process(self, tool_class: Type[xappt.BaseTool]):
        tool_name = tool_class.name()
        try:
            launch_command = self.launch_command(tool_name)
        except TypeError:
            self.critical(APP_TITLE, "Could not find executable")
        else:
            if platform.system() == "Windows":
                proc = subprocess.Popen(launch_command, creationflags=subprocess.CREATE_NEW_CONSOLE)
            else:
                proc = subprocess.Popen(launch_command)
            self.information(APP_TITLE, f"Launched {tool_name} (pid {proc.pid})")

    def selection_changed(self):
        help_text = ""
        selected_items = self.treeTools.selectedItems()
        if len(selected_items):
            tool_class = selected_items[0].data(0, self.ROLE_TOOL_CLASS)  # type: xappt.BaseTool
            if tool_class is not None and len(tool_class.help()):
                help_text = f"{tool_class.name()}: {tool_class.help()}"
        self.labelHelp.setText(help_text)

    def _filter_key_press(self, event: QtGui.QKeyEvent):
        if event.key() == QtCore.Qt.Key_Escape:
            self.txtSearch.clear()
        self.__txtSearch_keyPressEvent_orig(event)

    def on_filter_tools(self, text: str):
        if len(text) == 0:
            # noinspection PyTypeChecker
            iterator = QtWidgets.QTreeWidgetItemIterator(self.treeTools, QtWidgets.QTreeWidgetItemIterator.All)
            while iterator.value():
                item = iterator.value()
                item.setHidden(False)
                iterator += 1
            return
        search_terms = [item.lower() for item in text.split(" ") if len(item)]
        self._filter_branch(search_terms, self.treeTools.invisibleRootItem())

    def _filter_branch(self, search_terms: List[str], parent: QtWidgets.QTreeWidgetItem) -> int:
        visible_children = 0
        for c in range(parent.childCount()):
            child = parent.child(c)
            item_type = child.data(0, self.ROLE_ITEM_TYPE)
            if item_type == self.ITEM_TYPE_TOOL:
                child_text = child.text(0).lower()
                child_help = child.toolTip(0).lower()
                visible_children += 1
                item_hidden = False
                for term in search_terms:
                    if term not in child_text and term not in child_help:
                        item_hidden = True
                        break
                child.setHidden(item_hidden)
                if item_hidden:
                    visible_children -= 1
            elif item_type == self.ITEM_TYPE_COLLECTION:
                visible_children += self._filter_branch(search_terms, child)
            else:
                raise NotImplementedError
        parent.setHidden(visible_children == 0)
        return visible_children



#!/usr/bin/env python3

import argparse
import os
import platform
import re
import shutil
import sys
import venv

from distutils import sysconfig
from typing import Generator, List, Optional

import xappt

REQUIRED_PACKAGES = ['PyQt5==5.15.2', 'xappt==0.5.0']

SSH_REGEX = re.compile(r"^(?P<user>[^:]+?)(?::(?P<pass>[^/].*?))?@(?P<host>.*?):(?:(?P<port>\d+)/)?"
                       r"(?P<path>.*?/.*?)$", re.I)
URL_REGEX = re.compile(r"^(?P<protocol>.*?)://(?:(?P<user>.*?)(?::(?P<pass>.*?))?@)?"
                       r"(?P<domain>.*?)(?::(?P<port>\d+))?/(?P<path>.*)$", re.I)

if platform.system() == "Windows":
    VENV_BIN = "Scripts"
    PYTHON_EXT = ".exe"
else:
    VENV_BIN = "bin"
    PYTHON_EXT = ""

PYQT5_INIT_WINDOWS = """
__path__ = __import__('pkgutil').extend_path(__path__, __name__)


def find_qt():
    import os, sys

    qtcore_dll = 'Qt5Core.dll'

    path_var = os.environ['PATH']

    for path in path_var.split(os.pathsep):
        if os.path.isfile(os.path.join(path, qtcore_dll)):
            os.add_dll_directory(path)
            return

    search_paths = [
        os.getcwd(),
        os.path.join(os.path.dirname(__file__), "Qt", "bin"),
    ]

    if sys.executable is not None:
        python_dir = os.path.dirname(sys.executable)
        search_paths.append(python_dir)

    for path in search_paths:
        if os.path.isfile(os.path.join(path, qtcore_dll)):
            os.environ['PATH'] = path + os.pathsep + path_var
            os.add_dll_directory(path)
            return


find_qt()
del find_qt
"""


def build_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser()

    parser.add_argument('-s', '--source', help='A git URL to clone, or folder to copy, for the build. If not '
                                               'specified the current code base will be used.')
    parser.add_argument('-b', '--branch', default="master", help='The branch to check out, if source is a git URL')
    parser.add_argument('-o', '--output', help='A folder that will contain the built program.')
    parser.add_argument('-p', '--plugins', action='append', help='Include an external xappt plugins folder or '
                                                                 'git url in the build.')
    parser.add_argument('-t', '--title', help='A custom window title for xappt_qt')
    parser.add_argument('-c', '--console', action='store_true', help='Show the console on Windows')

    return parser


def check_build_requirements():
    if shutil.which("git") is None:
        raise SystemExit("git was not found")
    try:
        import venv
    except ImportError:
        raise SystemExit("venv python module not found")


def resolve_package_string(package_str: str, *, strip_version: bool = False) -> Optional[str]:
    package_str = package_str.strip()

    # ignore blank lines
    if not len(package_str):
        return None

    # ignore comment lines
    if package_str[0] == "#":
        return None

    # strip comments
    package_str = package_str.replace("\t", " ")
    package_str = package_str.split(" #")[0].strip()

    if strip_version:
        for op in ('~=', '==', '!=', '<=', '>=', '<', '>'):  # see PEP 440
            package_str = package_str.split(op)[0].strip()

    if len(package_str):
        return package_str


def packages_from_requirements(req_file_path: str) -> Generator[str, None, None]:
    with open(req_file_path, "r") as fp:
        packages = fp.readlines()
    for package in packages:
        package = resolve_package_string(package)
        if package is not None:
            yield package


class Builder:
    def __init__(self, *, work_path: str):
        self.work_path = work_path
        self.cmd = xappt.CommandRunner(cwd=work_path)
        self.python_bin = None
        self.site_packages = None

    def create_venv(self):
        venv_path = os.path.join(self.work_path, "venv")
        venv.create(venv_path)

        self.python_bin = os.path.join(venv_path, VENV_BIN, 'python' + PYTHON_EXT)
        self.site_packages = sysconfig.get_python_lib(prefix=venv_path)

        install_command = (sys.executable, '-m', 'pip', 'install', 'pip', '-t', self.site_packages)
        if self.cmd.run(install_command, silent=False).result != 0:
            raise SystemExit("Could not install pip into virtual environment")

        self.cmd.env_path_prepend('PATH', os.path.dirname(self.python_bin))
        self.cmd.env_path_prepend('PATH', os.path.join(self.site_packages, 'bin'))
        self.cmd.env_path_prepend('PYTHONPATH', self.site_packages)
        self.cmd.env_var_add('VIRTUAL_ENV', venv_path)
        self.cmd.env_var_remove('PYTHONHOME')

    def install_python_package(self, package_name: str):
        install_command = (self.python_bin, '-m', 'pip', 'install', package_name, '-t', self.site_packages)
        if self.cmd.run(install_command, silent=False).result != 0:
            raise SystemExit(f"Error installing {package_name}")

    def install_python_requirements(self, req_file_path: str, *, exclude: Optional[List[str]] = None):
        if exclude is not None:
            for package in packages_from_requirements(req_file_path):
                package_name = resolve_package_string(package, strip_version=True)
                if package_name in exclude:
                    continue
                self.install_python_package(package)
        else:
            # no exclusions - just a standard requirements.txt install
            install_command = (self.python_bin, '-m', 'pip', 'install', '-r', req_file_path,
                               '-t', self.site_packages)
            if self.cmd.run(install_command, silent=False).result != 0:
                raise SystemExit(f"Error installing {req_file_path}")

    def clone_repository(self, url: str, *, destination: str,  branch: Optional[str] = None):
        if branch is not None:
            git_command = ("git", "clone", "-b", branch, url, destination)
        else:
            git_command = ("git", "clone", url, destination)
        if self.cmd.run(git_command, silent=False).result != 0:
            raise SystemExit(f"Clone of {url} failed")

    def clone_or_copy_repository(self, url: str, *, destination: str, **kwargs) -> str:
        repo_name = os.path.basename(url)
        repo_dest = os.path.join(destination, repo_name)
        if os.path.isdir(url):
            shutil.copytree(url, repo_dest)
            return repo_dest
        else:
            for regex in (SSH_REGEX, URL_REGEX):
                match = regex.match(url)
                if match is not None:
                    self.clone_repository(url=url, destination=repo_dest, **kwargs)
                    return repo_dest
        raise NotImplementedError


def patch_pyqt5(*, site_packages: str):
    if platform.system() != "Windows":
        return
    pyqt_init_path = os.path.join(site_packages, "PyQt5", "utilities/__init__.py")
    assert os.path.isfile(pyqt_init_path)
    os.rename(pyqt_init_path, f"{pyqt_init_path}.bak")
    with open(pyqt_init_path, "w") as fp:
        fp.write(PYQT5_INIT_WINDOWS)


def get_version(version_path: str) -> str:
    with open(version_path, "r") as fp:
        version_contents = fp.read()

    loc = locals()
    exec(version_contents, {}, loc)
    __version__ = loc['__version__']
    assert __version__ is not None

    return __version__


def update_build(version_path: str, new_build: str):
    version = get_version(version_path)
    with open(version_path, "w") as fp:
        fp.write(f'__version__ = "{version}"\n')
        fp.write(f'__build__ = "{new_build}"\n')


def update_app_title(constants_path: str, *, new_title: str):
    with open(constants_path, "r") as fp:
        lines = fp.readlines()

    with open(constants_path, "w") as fp:
        replaced_title = False
        for line in lines:
            if line.startswith('APP_TITLE = '):
                line = f'APP_TITLE = "{new_title}"\n'
                replaced_title = True
            fp.write(line)

    if not replaced_title:
        raise SystemExit(f"APP_TITLE declaration not fount in '{constants_path}'")


def find_qt() -> str:
    import PyQt5

    if platform.system() == "Linux":
        qt5_bin = "lib"
        qt5core_lib = "libQt5Core.so.5"
    elif platform.system() == "Windows":
        qt5_bin = "bin"
        qt5core_lib = "Qt5Core.dll"
    else:
        raise NotImplementedError

    bin_path = os.path.join(os.path.dirname(PyQt5.__file__), "Qt", qt5_bin)
    if os.path.isfile(os.path.join(bin_path, qt5core_lib)):
        return bin_path

    raise FileNotFoundError("Could not find PyQt5's QtCore library")


def inject_plugin_import(plugin_path: str, *, target_file: str, line_num: int):
    with open(target_file, "r") as fp:
        lines = fp.readlines()

    prefix = "".join(lines[:line_num])
    suffix = "".join(lines[line_num:])

    plugin_name = os.path.splitext(os.path.basename(plugin_path))[0]

    import_lines = ""
    for plugin in (plugin_name, f"{plugin_name}.plugins"):
        import_lines += f"\ntry:\n    import {plugin}\nexcept ImportError:\n    pass\n\n"

    with open(target_file, "w") as fp:
        fp.write(prefix)
        fp.write(import_lines)
        fp.write(suffix)


def main(args) -> int:
    check_build_requirements()

    parser = build_parser()
    options = parser.parse_args(args=args)

    output_path = os.path.abspath(options.output)
    if os.path.isdir(output_path):
        raise SystemExit(f"Error, output path exists: {output_path}")

    with xappt.temp_path() as tmp:
        builder = Builder(work_path=tmp)
        builder.create_venv()

        if options.source is not None:
            repo_path = builder.clone_or_copy_repository(options.source, destination=tmp, branch=options.branch)
            req_path = os.path.join(repo_path, "requirements.txt")
            builder.install_python_requirements(req_path)
            version_path = os.path.join(repo_path, 'xappt_qt', '__version__.py')
            commit_id = xappt.git_tools.commit_id(repo_path, short=True)
            update_build(version_path, commit_id)
        else:
            repo_path = os.path.join(tmp, "xappt_qt")
            root_path = os.path.dirname(os.path.dirname(__file__))
            shutil.copytree(root_path, repo_path)
            for package in REQUIRED_PACKAGES:
                builder.install_python_package(package)

        entry_point = os.path.join(repo_path, "xappt_qt", "main.py")
        assert os.path.isfile(entry_point)

        plugins_destination = os.path.join(tmp, "plugins")
        for plugin in options.plugins or []:  # in case options.plugins is None
            plugin_path = builder.clone_or_copy_repository(plugin, destination=plugins_destination)
            req_file = os.path.join(plugin_path, "requirements.txt")
            if os.path.isfile(req_file):
                builder.install_python_requirements(req_file, exclude=["xappt", "xappt-qt"])
            inject_plugin_import(plugin_path, target_file=entry_point, line_num=8)
            builder.cmd.env_path_prepend("PYTHONPATH", plugin_path)

        if options.title is not None:
            constants_path = os.path.join(repo_path, "xappt_qt", "constants.py")
            update_app_title(constants_path, new_title=options.title)

        builder.cmd.env_path_prepend("PATH", find_qt())

        patch_pyqt5(site_packages=builder.site_packages)

        nuitka_package = "nuitka==0.6.9.2"
        builder.install_python_package(nuitka_package)

        nuitka_command = ["python", "-m", "nuitka", "--standalone", "--recurse-all",
                          f"--windows-icon={os.path.join(repo_path, 'xappt_qt', 'resources', 'icons', 'appicon.ico')}",
                          "--plugin-enable=qt-plugins", f"--output-dir={output_path}"]

        if not options.console:
            nuitka_command.append("--windows-disable-console")

        nuitka_command.append(entry_point)

        builder.cmd.run(nuitka_command, silent=False)

    return 0


if __name__ == '__main__':
    if sys.version_info[:2] < (3, 7):
        raise SystemExit("Python 3.7 or higher is required.")
    sys.exit(main(sys.argv[1:]))



from typing import Any, Callable, Dict, Generator, List, Optional, Type

from PyQt5 import QtWidgets, QtCore

import xappt

from xappt_qt.gui.widgets import *
from xappt_qt.gui.delegates import SimpleItemDelegate


class ToolPage(QtWidgets.QWidget):
    def __init__(self, tool: xappt.BaseTool, parent: Optional[QtWidgets.QWidget] = None):
        super().__init__(parent=parent)

        self.convert_dispatch: Dict[Type, Callable] = {
            int: self._convert_int,
            bool: self._convert_bool,
            float: self._convert_float,
            str: self._convert_str,
            list: self._convert_list,
        }

        self.tool = tool
        self.build_ui()

    # noinspection PyAttributeOutsideInit
    def build_ui(self):
        self.grid = QtWidgets.QGridLayout()
        self.grid.setColumnStretch(0, 0)
        self.grid.setColumnStretch(1, 1)
        self.grid.setColumnStretch(0, 0)
        self.grid.setHorizontalSpacing(16)
        self.grid.setVerticalSpacing(8)

        self.setLayout(self.grid)
        self._load_tool_parameters()

    @staticmethod
    def get_caption(param: xappt.Parameter) -> str:
        caption_default = param.name.replace("_", " ").title()
        caption = param.options.get("caption", caption_default)
        return caption

    def update_tool_choices(self, param: xappt.Parameter):
        """ Given that multiple parameter types can be updated at runtime,
        it's easier just to remove and recreate the widget rather than
        reimplementing a lot of the same functionality to update existing
        widgets. """
        widget = param.metadata.get('widget')
        if widget is None:
            return

        # find and remove existing widget
        index = self.grid.indexOf(widget)
        row, column, *_ = self.grid.getItemPosition(index)
        self.grid.takeAt(index)
        widget.deleteLater()
        param.on_choices_changed.clear()

        # create a new widget to replace it
        new_widget = self.convert_parameter(param)
        self.grid.addWidget(new_widget, row, column)
        param.metadata['widget'] = new_widget
        param.on_choices_changed.add(self.update_tool_choices)

    def _load_tool_parameters(self):
        for i, param in enumerate(self.tool.parameters()):
            label = QtWidgets.QLabel(self.get_caption(param))
            label.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)
            label.setToolTip(param.description)
            param.metadata['label'] = label
            self.grid.addWidget(label, i, 0)

            error = ErrorLabel()
            param.metadata['error'] = error
            self.grid.addWidget(error, i, 2)

            widget = self.convert_parameter(param)
            widget.setToolTip(param.description)
            param.metadata['widget'] = widget  # this lets us avoid lambdas
            self.grid.addWidget(widget, i, 1)

            self.widget_options_updated(param)

            param.on_choices_changed.add(self.update_tool_choices)
            param.on_value_changed.add(self.widget_value_updated)
            param.on_options_changed.add(self.widget_options_updated)

    def convert_parameter(self, param: xappt.Parameter) -> QtWidgets.QWidget:
        convert_fn = self.convert_dispatch.get(param.data_type)
        if convert_fn is not None:
            return convert_fn(param)
        return QtWidgets.QWidget()

    def _convert_int(self, param: xappt.Parameter) -> QtWidgets.QWidget:
        if param.choices is not None:
            return self._convert_int_choice(param)
        else:
            return self._convert_int_spin(param)

    def _convert_int_choice(self, param: xappt.Parameter) -> QtWidgets.QWidget:
        w = QtWidgets.QComboBox()
        w.addItems(param.choices)
        for v in (param.value, param.default):
            if v is not None:
                if isinstance(v, str):
                    if v in param.choices:
                        index = w.findText(v)
                        w.setCurrentIndex(index)
                elif isinstance(v, int):
                    if 0 <= v < w.count():
                        w.setCurrentIndex(v)
                break
        else:
            param.value = w.currentIndex()

        if param.options.get("searchable"):
            self.setup_combo_completer(w)

        w.currentIndexChanged[str].connect(lambda x: self.update_tool_param(param.name, x))
        param.metadata['ui-setter'] = w.setCurrentIndex
        return w

    def _convert_int_spin(self, param: xappt.Parameter) -> QtWidgets.QWidget:
        w = QtWidgets.QSpinBox(parent=self)
        minimum = param.options.get("minimum", -999999999)
        maximum = param.options.get("maximum", 999999999)
        w.setMinimum(minimum)
        w.setMaximum(maximum)
        for v in (param.value, param.default):
            if v is not None:
                w.setValue(v)
                break
        else:
            param.value = w.value()
        w.valueChanged[int].connect(lambda x: self.update_tool_param(param.name, x))
        param.metadata['ui-setter'] = w.setValue
        return w

    def _convert_bool(self, param: xappt.Parameter) -> QtWidgets.QWidget:
        w = QtWidgets.QCheckBox()
        for v in (param.value, param.default):
            if v is not None:
                w.setChecked(v)
                break
        else:
            param.value = w.isChecked()
        w.stateChanged.connect(lambda x: self.update_tool_param(param.name, x == QtCore.Qt.Checked))
        param.metadata['ui-setter'] = w.setChecked
        return w

    def _convert_str(self, param: xappt.Parameter) -> QtWidgets.QWidget:
        if param.choices is not None:
            return self._convert_str_choice(param)
        else:
            return self._convert_str_edit(param)

    def _convert_str_choice(self, param: xappt.Parameter) -> QtWidgets.QWidget:
        w = QtWidgets.QComboBox()
        w.addItems(param.choices)
        for v in (param.value, param.default):
            if v is not None:
                if v in param.choices:
                    index = w.findText(param.default)
                    w.setCurrentIndex(index)
                    break
        else:
            param.value = w.currentText()

        if param.options.get("searchable"):
            self.setup_combo_completer(w)

        w.currentIndexChanged[str].connect(lambda x: self.update_tool_param(param.name, x))
        param.metadata['ui-setter'] = lambda s, widget=w: widget.setCurrentIndex(widget.findText(s))
        return w

    def _convert_str_edit(self, param: xappt.Parameter) -> QtWidgets.QWidget:
        ui = param.options.get("ui")
        if ui == "folder-select":
            w = FileEdit(mode=FileEdit.MODE_CHOOSE_DIR)
            w.onSetFile.connect(lambda x: self.update_tool_param(param.name, x))
        elif ui == "file-open":
            w = FileEdit(accept=param.options.get("accept"), mode=FileEdit.MODE_OPEN_FILE)
            w.onSetFile.connect(lambda x: self.update_tool_param(param.name, x))
        elif ui == "file-save":
            w = FileEdit(accept=param.options.get("accept"), mode=FileEdit.MODE_SAVE_FILE)
            w.onSetFile.connect(lambda x: self.update_tool_param(param.name, x))
        else:
            w = QtWidgets.QLineEdit()
            w.textChanged.connect(lambda x: self.update_tool_param(param.name, x))

        for v in (param.value, param.default):
            if v is not None:
                w.setText(param.value)
                break
        else:
            w.setText("")
        param.metadata['ui-setter'] = w.setText
        return w

    # noinspection DuplicatedCode
    def _convert_float(self, param: xappt.Parameter) -> QtWidgets.QWidget:
        w = QtWidgets.QDoubleSpinBox(parent=self)
        minimum = param.options.get("minimum", -999999999.0)
        maximum = param.options.get("maximum", 999999999.0)
        w.setMinimum(minimum)
        w.setMaximum(maximum)
        if param.default is not None:
            w.setValue(param.default)
        param.value = w.value()
        w.valueChanged[float].connect(lambda x: self.update_tool_param(param.name, x))
        param.metadata['ui-setter'] = w.setValue
        return w

    def _convert_list(self, param: xappt.Parameter) -> QtWidgets.QWidget:
        w = CheckList(searchable=param.options.get('searchable', False))
        if param.choices is not None:
            w.add_items(param.choices)
        for v in (param.value, param.default):
            if v is not None:
                w.check_items(v)
                break
        param.value = list(w.checked_items())
        w.item_changed.connect(lambda: self.update_list_param(param.name))
        param.metadata['ui-setter'] = lambda value, widget=w: self.set_list_value(value, widget)
        return w

    def update_list_param(self, name: str):
        param: xappt.Parameter = getattr(self.tool, name)
        widget: CheckList = param.metadata.get('widget')
        if widget is None:
            return
        checked_items = list(widget.checked_items())
        param.on_value_changed.paused = True
        param.value = param.validate(checked_items)
        param.on_value_changed.paused = False

    def update_tool_param(self, name: str, value: Any):
        param: xappt.Parameter = getattr(self.tool, name)
        error: ErrorLabel = param.metadata['error']
        try:
            param.value = param.validate(value)
        except xappt.ParameterValidationError as e:
            error.set_error(str(e))
        else:
            error.reset()

    @staticmethod
    def set_list_value(items: List[str], widget: CheckList):
        widget.uncheck_all()
        widget.check_items(items)

    @staticmethod
    def widget_value_updated(param: xappt.Parameter):
        setter = param.metadata.get('ui-setter')
        if setter is None:
            return
        setter(param.value)

    @staticmethod
    def widget_options_updated(param: xappt.Parameter):
        label: Optional[QtWidgets.QLabel] = param.metadata.get('label')
        widget: Optional[QtWidgets.QWidget] = param.metadata.get('widget')
        error: Optional[ErrorLabel] = param.metadata.get('error')

        visible = param.option("visible", True)
        if label is not None:
            label.setVisible(visible)
        if widget is not None:
            widget.setVisible(visible)
        if error is not None:
            error.setVisible(visible)

        enabled = param.option("enabled", True)
        if widget is not None:
            widget.setEnabled(enabled)

    def disconnect(self):
        for param in self.tool.parameters():
            param.on_value_changed.clear()
            param.on_options_changed.clear()
            param.on_choices_changed.clear()

    @staticmethod
    def setup_combo_completer(combo_widget: QtWidgets.QComboBox):
        combo_widget.setEditable(True)
        combo_widget.setInsertPolicy(QtWidgets.QComboBox.NoInsert)
        combo_widget.lineEdit().editingFinished.connect(
            lambda: combo_widget.setCurrentIndex(combo_widget.findText(combo_widget.currentText())))
        completer = combo_widget.completer()
        completer.setCompletionMode(QtWidgets.QCompleter.PopupCompletion)
        completer.popup().setAlternatingRowColors(True)
        completer.setCaseSensitivity(QtCore.Qt.CaseInsensitive)
        completer.setFilterMode(QtCore.Qt.MatchContains)

